/*
 *	dummyfill.c
 *
 *	Example program for v4l loopback device that works with v4l/v4l2.
 *	Based on the programs "invert" and "dummy" from the author of
 *	 "v4l loopback device" Jeroen Vreeken.
 *
 *	This program feeds a v4l loopback device with squares,
 *	in RGB24 and RGB565 format, trying to generate some kind of motion.
 *
 *	Copyright 2004, 2005 by Oliver Gonzalez (mallok25@gmail.com)
 *
 *	This software is distributed under the GNU public license version 2
 *	See also the file 'COPYING'.
 *
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <linux/videodev.h>

#define byte unsigned char
#define uint unsigned int

#define RGB24 3
#define RGB16 2
#define SPEED 0.2f
#define GROW 0.01f
#define TRUE 1
#define FALSE 0
#define FORWARD SPEED
#define BACKWARD -SPEED
#define BIGGER GROW
#define SMALLER -GROW
#define SQUARE_BIG_MAX 100
#define SQUARE_BIG_MIN 10
#define BACKGROUND_COLOR 0
#define SCREEN_MAX_SQUARES 4
#define SCREEN_MIN_SQUARES 2

struct square {
    float x_pos;
    float y_pos;
    float big;
    float big_inc;
    float x_direction;
    float y_direction;
    uint colour;
};

static byte DEBUG;
static uint FPS;
static uint PIXEL_FORMAT;

void initialize_random_seed () {
    uint seed;
    int handle;
    
    if ((handle = open("/dev/urandom", O_RDONLY | O_NONBLOCK, 0)) > -1) {
	  // Uses random device (better)
	  read (handle, &seed, sizeof(uint));
	  close (handle);
    } else {
	  // Uses the system local time
	  struct timeval tv;
	  struct timezone tz;
	  gettimeofday (&tv, &tz);
	  seed = tv.tv_usec;
    }
    srandom(seed);
}

uint get_random (uint max, uint min) {
    uint value;
    while ((value = random() % max) <= min);
    
    return value;
}

int start_pipe (int dev, uint width, uint height)
{
    struct video_capability vid_caps;
    struct video_window vid_win;
    struct video_picture vid_pic;

    // Gets video device capabilities
    if (ioctl (dev, VIDIOCGCAP, &vid_caps) == -1) {
	  perror ("ioctl (VIDIOCGCAP)");
	  return (1);
    }
    
    // Gets video device picture formats
    if (ioctl (dev, VIDIOCGPICT, &vid_pic) == -1) {
	  perror ("ioctl VIDIOCGPICT");
	  return (1);
    }
    
    // Sets video device format
    if (PIXEL_FORMAT == RGB24)
	  vid_pic.palette = VIDEO_PALETTE_RGB24;
    else
  	  if (PIXEL_FORMAT == RGB16)
		vid_pic.palette = VIDEO_PALETTE_RGB565;
    
    if (ioctl (dev, VIDIOCSPICT, &vid_pic) == -1) {
	  perror ("ioctl VIDIOCSPICT");
	  return (1);
    }
    
    // Gets video device capture sizes
    if (ioctl (dev, VIDIOCGWIN, &vid_win) == -1) {
	  perror ("ioctl VIDIOCGWIN");
	  return (1);
    }
    
    vid_win.width = width;
    vid_win.height = height;
    
    // Sets video device capture size
    if (ioctl (dev, VIDIOCSWIN, &vid_win) == -1) {
	  perror ("ioctl VIDIOCSWIN");
	  return (1);
    }
	
    return 0;
}

// Draws a horizontal line
inline void drawline_horizontal (byte *map, uint x_init, uint y_init,
		uint x_final, uint y_final, uint width, uint colour) {

    // Decodes the colour
    byte *rgb = (byte *) &colour;

    int i, t, dy, offset;
    
    if (x_init > x_final) {
	  t = x_init; x_init = x_final; x_final = t;
    }
    if (y_init > y_final) {
	  t = y_init; y_init = y_final; y_final = t;
    }

    dy = y_final - y_init;
    
    /* horizontal line */
    if (dy == 0) { 
	  int y_offset = y_init * width;
      offset = (y_offset + x_init) * PIXEL_FORMAT;
	
	  for (i = x_init; i < x_final; i++) {
	    // Perfomance tweak
	    if (PIXEL_FORMAT == RGB16) {
		  map[offset + 1] = rgb[0];
		  map[offset + 2] = rgb[1];
	    }
	    else
	    if (PIXEL_FORMAT == RGB24) {
		  map[offset + 1] = rgb[0];
		  map[offset + 2] = rgb[1];
		  map[offset + 3] = rgb[2];
	    }
	    
	    offset += PIXEL_FORMAT;
	}
	    
        return;
    }
}

void draw_square (byte *map, uint x_init, uint y_init,
		uint x_final, uint y_final, uint width, uint colour) {

    // Draws a square by drawing lines
    int y;
    for (y = y_init; y < y_final; y++) 
	  drawline_horizontal (map, x_init, y, x_final, y, width, colour);
}

void next_frame (byte *map, uint width, uint height, struct square *squares,\
								uint n_squares)
{
    uint x, y, big, skipped_fps;
    int i, direction_switched;
    
    // Sets the background to black
    memset (map, BACKGROUND_COLOR, width * height * PIXEL_FORMAT);

    // Draws squares
    for (i = 0; i < n_squares; i++) {
	  x = (uint) squares[i].x_pos;
	  y = (uint) squares[i].y_pos;
	  big = (uint) squares[i].big;

	  // Checks boundaries
	  if (big >= SQUARE_BIG_MAX) {
		// Switches square direction
		if ((get_random(2,0)) == 1)
		  squares[i].x_direction = FORWARD;
		else
		  squares[i].x_direction = BACKWARD;

		squares[i].big_inc = SMALLER;
		big = SQUARE_BIG_MAX;
	  }
	  else if (big <= SQUARE_BIG_MIN) {
		// Switches square direction
		if ((get_random(2,0)) == 1)
		  squares[i].y_direction = FORWARD;
		else
		  squares[i].y_direction = BACKWARD;

		squares[i].big_inc = BIGGER;
		big = SQUARE_BIG_MIN;
	  }

	  direction_switched = FALSE;
	
	  if (x >= (width - big)) {
		squares[i].x_direction = BACKWARD;
		x = (width - big);
		direction_switched = TRUE;
	  }
	  else if (x <= 1) {
		squares[i].x_direction = FORWARD;
		x = 1;
		direction_switched = TRUE;
	  }
	    
	  if (y >= (height - big)) {
		squares[i].y_direction = BACKWARD;
		y = (height - big);
		direction_switched = TRUE;
	  }
	  else if (y <= 1) {
		squares[i].y_direction = FORWARD;
		y = 1;
		direction_switched = TRUE;
	  }
	    
	  // Colour changing
	  if (direction_switched == TRUE)
		squares[i].colour++; 

	  // Draws the square
	  draw_square (map, x, y, x + big,\
		     y + big, width, squares[i].colour);

	  // Calculating the skipped FPS since last frame
	  skipped_fps = 25 - FPS;
	
	  // Moves to any direction
	  squares[i].x_pos += (squares[i].x_direction * skipped_fps);
	  squares[i].y_pos += (squares[i].y_direction * skipped_fps);
	  squares[i].big += (squares[i].big_inc * skipped_fps);
    }
}
		 
int put_image (int dev, byte *image, uint width, uint height)
{
    // Writes the image to the v4l device
    if (write (dev, image, width * height * PIXEL_FORMAT)\
			!= (width * height * PIXEL_FORMAT)) {
	  perror("Error writing image to pipe!");
	  return 0;
    }
    
    return 1;
}


void fill_squares_data (struct square *squares, uint n_squares, uint width,\
								uint height) {

    int i;
    for (i = 0; i < n_squares; i++) {
	  squares[i].big = (float) get_random(SQUARE_BIG_MAX, SQUARE_BIG_MIN);
	  squares[i].x_pos = (float) get_random(width - (uint) squares[i].big, 5);
	  squares[i].y_pos = (float) get_random(height - (uint) squares[i].big, 5);
	  squares[i].colour = get_random(0xFFFFFF, 0x0);
	
	if ((get_random(2, 0)) == 1)
	  squares[i].big_inc = FORWARD;
	else
	  squares[i].big_inc = BACKWARD;
	    
	if ((get_random(2, 0)) == 1)
	  squares[i].x_direction = FORWARD;
	else
	  squares[i].x_direction = BACKWARD;

	if ((get_random(2, 0)) == 1)
	  squares[i].y_direction = FORWARD;
	else
	  squares[i].y_direction = BACKWARD;
    }
}

int main (int argc, char **argv)
{
    struct square *squares;
    byte *image;
    uint width, height;
    uint n_squares;
    int devout;
    
    // Debug is disabled for default
    DEBUG = FALSE;

    if (argc < 4) {
	printf("\nInsufficient parameters.\n");
	printf("Usage: dummyfill vloopback-device WidthxHeightxDepth FPS [-v]\n\n");
	printf("Example: dummyfill /dev/video0 352x288x24 25\n\n");
	
	exit(1);
    } else {
	// Check the verbose option
      if (argc == 5) {
		if (!strcmp(argv[4], "-v")) {
		  DEBUG = TRUE;
		
		  printf("dummyfill is being executed in verbose mode.\n");
		}
	  }
    }
    
    // Gets resolution data from command line arguments
    sscanf(argv[2], "%dx%dx%d", &width, &height, &PIXEL_FORMAT);
    sscanf(argv[3], "%d", &FPS);

    // If enters an incorrect color depth value, it exits
    if (PIXEL_FORMAT == 16) {
	  PIXEL_FORMAT = RGB16;
    }
    else
    if (PIXEL_FORMAT == 24) {
	  PIXEL_FORMAT = RGB24;
    } else {
	  perror ("Invalid colour depth value, only 16 and 24 bits supported.");
	  exit(1);
    }
    
    // If enters an incorrect FPS value, it exits
    if (FPS > 25 || FPS < 1) {
	  perror ("Invalid FPS value, range is from 1 to 25.");
	  exit(1);
    }
    
    if (DEBUG)
	  printf("Device: %s\nResolution: %dx%d\nColor depth: %d\nFPS: %d\n",\
				argv[1], width, height, PIXEL_FORMAT, FPS);
    
    image = (byte *) malloc(height * width * PIXEL_FORMAT);

    // Opens the device
    devout = open(argv[1], O_RDWR | O_NONBLOCK, 0);
    if (devout < 0) {
	  perror ("Failed to open the specified device.");
	  exit(1);
    }
    
    if (DEBUG)
      printf("Deviced opened.\n");

    // Opens the output device
    if (start_pipe(devout, width, height)) {
	  perror("Failed to initialize output pipe, exiting...");
	  exit(1);
    }
    
    if (DEBUG)
	  printf("Pipe opened.\n");

    // Initializes random seed
    initialize_random_seed();
    
    // Allocates a random number of squares
    n_squares = get_random(SCREEN_MAX_SQUARES, SCREEN_MIN_SQUARES);
    squares = (struct square *) malloc(sizeof(struct square) * n_squares);
    
    if (DEBUG)
      printf("Numbers of squares: %d\n", n_squares);
    
    // Fills them
    fill_squares_data(squares, n_squares, width, height);
        
    if (DEBUG)
	  printf("Video stream started.\n");
	
    while (1) {
	  next_frame(image, width, height, squares, n_squares);
	  if (put_image(devout, image, width, height) == 0)
		exit(1);
		
	  usleep(1000000 / FPS);
    }

    // Exits
    close(devout);
    free(image);
    
    exit(0);
}
